"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const _ = require("lodash");
const metadata_1 = require("../metadata");
const decorators_1 = require("../decorators");
class DbImpl {
    constructor(modelFn, model) {
        this.modelFn = modelFn;
        this.model = model;
        this.createCopyValsFn();
        this.createTransformQuery();
    }
    create(t) {
        return __awaiter(this, void 0, void 0, function* () {
            if (t instanceof Array) {
                t = _.cloneDeep(t);
                t.forEach((t, idx, arr) => { arr[idx] = this.transformQueryWhere(t); });
                let results = yield this.model.bulkCreate(t);
                // return this.wrapResults(results);
                return true;
            }
            else {
                t = this.transformQueryWhere(t);
                let result = yield this.model.create(t);
                return this.wrapResult(result);
            }
        });
    }
    findById(id, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (options)
                options = this.transformQuery(options);
            let result = yield this.model.findById(id, options);
            return result && this.wrapResult(result);
        });
    }
    findOne(query) {
        return __awaiter(this, void 0, void 0, function* () {
            query = this.transformQuery(query);
            let result = yield this.model.findOne(query);
            return result && this.wrapResult(result);
        });
    }
    findOrCreate(query, defaults) {
        return __awaiter(this, void 0, void 0, function* () {
            query = this.transformQueryWhere(query);
            defaults = this.transformQueryWhere(defaults);
            let [result, created] = yield this.model.findOrCreate({ where: query, defaults: defaults || {} });
            return [result && this.wrapResult(result), created];
        });
    }
    findAndCountAll(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (query)
                query = this.transformQuery(query);
            let results = yield this.model.findAndCountAll(query);
            return { count: results.count, results: this.wrapResults(results.rows) };
        });
    }
    findAll(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (query)
                query = this.transformQuery(query);
            let results = yield this.model.findAll(query);
            return this.wrapResults(results);
        });
    }
    all(query) {
        return this.findAll(query);
    }
    count(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (query)
                query = this.transformQuery(query);
            return yield this.model.count(query);
        });
    }
    max(field) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.model.max(field);
        });
    }
    min(field) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.model.min(field);
        });
    }
    sum(field) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.model.sum(field);
        });
    }
    save(t) {
        return __awaiter(this, void 0, void 0, function* () {
            let [result, created] = yield this.findOrCreate({ id: t.id }, t);
            return result;
        });
    }
    update(query, replace, returning = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else {
                query = this.transformQuery(query);
                if (returning) {
                    let results = yield this.model.findAll(query);
                    let ids = results.map((result) => result.id);
                    query = { where: { id: { $in: ids } } };
                }
            }
            replace = this.transformQueryWhere(replace);
            let [affected, results] = yield this.model.update(replace, query);
            if (returning)
                results = this.wrapResults(yield this.model.findAll(query));
            return [affected, results];
        });
    }
    updateOrCreate(query, defaults) {
        return __awaiter(this, void 0, void 0, function* () {
            let [result, created] = yield this.findOrCreate(query, defaults);
            if (!created) {
                let worked = yield this.update({ where: query }, defaults);
                if (!worked)
                    throw new Error("Failed to update or create a model.");
                let resultOrNull = yield this.findOne(query);
                if (!resultOrNull)
                    throw new Error("Updated row, but could not find it afterwards.");
                result = resultOrNull;
            }
            return [result, created];
        });
    }
    destroy(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else
                query = this.transformQuery(query);
            return yield this.model.destroy(query);
        });
    }
    isId(query) {
        return typeof query === 'number' || typeof query == 'string';
    }
    isT(query) {
        return !!query.id;
    }
    createCopyValsFn() {
        let allProps = [];
        let directTransformFn = (val) => val;
        var props = Reflect.getOwnMetadata(metadata_1.ModelPropertiesSym, this.modelFn.prototype) || [];
        for (var q = 0; q < props.length; q++) {
            var propName = props[q];
            var propMeta = Reflect.getOwnMetadata(metadata_1.PropMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model property metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let transformFn = directTransformFn;
            if (propMeta.type == decorators_1.Types.DATE)
                transformFn = (dateStr) => new Date(dateStr);
            allProps.push({ columnName: propMeta.columnName || propName, propertyName: propName, transformFn: transformFn });
        }
        // console.log(this.modelFn.name || this.modelFn, 'allProps:', `[\r\n    ${allProps.map(p => p.propertyName + ': ' + p.columnName).join(',\r\n    ')}\r\n]`);
        this.copyVals = function (sql, t) {
            for (var q = 0; q < allProps.length; q++) {
                // console.log(allProps[q].propertyName + ':', sql[allProps[q].propertyName]);
                // t[allProps[q].propertyName] = allProps[q].transformFn(sql[allProps[q].columnName]);
                t[allProps[q].propertyName] = allProps[q].transformFn(sql[allProps[q].propertyName]);
            }
            // console.log(JSON.stringify(sql));
        };
        //TODO: deep copy?
    }
    static getForeignModelDbImpl(foreignModel) {
        if (!foreignModel || !("db" in foreignModel))
            throw new Error("Cannot get DbImpl from ForeignModel that has not been resolved");
        let staticModel = foreignModel;
        return staticModel.db;
    }
    createTransformQuery() {
        let transforms = this.getTransforms();
        this.createTransformQueryWhere(transforms);
        this.createTransformQueryInclude(transforms);
        this.createTransformResult(transforms);
    }
    transformQuery(query) {
        let result = _.clone(query);
        if (query.where)
            result.where = this.transformQueryWhere(query.where);
        if (query.include)
            result.include = this.transformQueryInclude(query.include);
        return result;
    }
    getTransforms() {
        let transforms = [];
        let belongsTo = Reflect.getOwnMetadata(metadata_1.ModelBelongsToAssociationsSym, this.modelFn.prototype) || [];
        for (let q = 0; q < belongsTo.length; q++) {
            let propName = belongsTo[q];
            let propMeta = Reflect.getOwnMetadata(metadata_1.BelongsToMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model belongs-to metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let fkey = propMeta.foreignKey;
            if (fkey && typeof fkey !== 'string')
                fkey = fkey.name;
            if (!fkey)
                throw new Error(`Could not get foreign key for belongs-to property ${this.modelFn.name || this.modelFn}.${propName}`);
            transforms.push({
                type: 'belongs-to',
                columnName: fkey,
                fieldName: propName,
                foreignPkName: 'id',
                foreignDb: () => DbImpl.getForeignModelDbImpl(propMeta.foreignModel) //Has to be lazy to avoid race conditions
            });
        }
        let hasOne = Reflect.getOwnMetadata(metadata_1.ModelHasOneAssociationsSym, this.modelFn.prototype) || [];
        for (let q = 0; q < hasOne.length; q++) {
            let propName = hasOne[q];
            let propMeta = Reflect.getOwnMetadata(metadata_1.HasOneMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model has-one metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let fkey = propMeta.foreignKey;
            if (fkey && typeof fkey !== 'string')
                fkey = fkey.name;
            if (!fkey)
                throw new Error(`Could not get foreign key for has-one property ${this.modelFn.name || this.modelFn}.${propName}`);
            transforms.push({
                type: 'has-one',
                foreignColumnName: fkey,
                fieldName: propName,
                pkName: 'id',
                foreignDb: () => DbImpl.getForeignModelDbImpl(propMeta.foreignModel) //Has to be lazy to avoid race conditions
            });
        }
        return transforms;
    }
    createTransformQueryWhere(transforms) {
        this.transformQueryWhere = function (query) {
            if (!query)
                return query;
            query = _.clone(query);
            for (var q = 0; q < transforms.length; q++) {
                let transform = transforms[q];
                let fieldVal;
                switch (transform.type) {
                    case 'belongs-to':
                        fieldVal = query[transform.fieldName];
                        if (fieldVal) {
                            if (fieldVal[transform.foreignPkName])
                                fieldVal = fieldVal[transform.foreignPkName];
                            query[transform.columnName] = fieldVal;
                            delete query[transform.fieldName];
                        }
                        break;
                    case 'has-one':
                        fieldVal = query[transform.fieldName];
                        if (fieldVal) {
                            throw new Error(`Not implemented! Cannot include has-one value in where query`);
                        }
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
            return query;
        };
    }
    createTransformQueryInclude(transforms) {
        function getFieldModel(field) {
            for (let i = 0; i < transforms.length; i++) {
                if (field == transforms[i].fieldName)
                    return transforms[i].foreignDb().model;
            }
            return null;
        }
        this.transformQueryInclude = function (fields) {
            if (!fields)
                return fields;
            return fields.map(field => {
                let model = getFieldModel(field);
                if (!model)
                    throw new Error(`Cannot find field ${field} from include query`);
                return { model: model, as: field };
            });
        };
    }
    createTransformResult(transforms) {
        this.transformResult = function (sql, result) {
            result = _.clone(result);
            for (var q = 0; q < transforms.length; q++) {
                let transform = transforms[q];
                let foreignDb = transform.foreignDb();
                switch (transform.type) {
                    case 'belongs-to':
                        if (sql[transform.fieldName]) {
                            let t = new foreignDb.modelFn();
                            foreignDb.copyVals(sql[transform.fieldName], t);
                            result[transform.fieldName] = transform.foreignDb().transformResult(sql[transform.fieldName], t);
                        }
                        else if (sql[transform.columnName]) {
                            result[transform.fieldName] = sql[transform.columnName];
                            delete result[transform.columnName];
                        }
                        break;
                    case 'has-one':
                        if (sql[transform.fieldName]) {
                            let t = new foreignDb.modelFn();
                            foreignDb.copyVals(sql[transform.fieldName], t);
                            result[transform.fieldName] = foreignDb.transformResult(sql[transform.fieldName], t);
                        }
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
            return result;
        };
    }
    fromJson(json) {
        return this.wrapResult(json);
    }
    wrapResult(result) {
        let t = new this.modelFn();
        this.copyVals(result, t);
        return this.transformResult(result, t);
    }
    wrapResults(results) {
        return results.map(result => this.wrapResult(result));
    }
}
exports.DbImpl = DbImpl;
//# sourceMappingURL=db-impl.js.map